#include <ESP8266WiFi.h>
#include <PubSubClient.h>


//Länge Aufnahme 10000*100µs = 1sec, Größe Bytearray 2*SAMPLES (1  Messwert momentan noch 2 Bytes)
#define SAMPLES                10000
//Größe der einzelnen MQTT Pakete 
#define MQTTPACSIZE             500                
//Zeitspanne zwischen Messungen
#define MICRODELAY 100

const char* mqtt_server = "192.168.137.1";
const char* ssid = "IoTGruppeXY";
const char* password = "IoTWerkstatt";

//Steuerungsparameter über Seriellen Monitor 
int controlByte = 0;
//Laufvariable für Timer etc
int currSample = 0;
//Zwischenspeicher für Messwerte
int wert = 0;
byte dataCollection[SAMPLES*2];
byte mqttpack[MQTTPACSIZE];

//-------------- definition mqtt-object ueber WiFi
WiFiClient   espClient; 
PubSubClient mqttclient(espClient);

//--------- list of mqtt callback functions 
#define MAX_MQTT_SUB 10 // maximal 10 subscriptions erlaubt
typedef void (*mqtthandle) (byte*,unsigned int);
typedef struct {       // Typdeklaration Callback
  String topic;        // mqtt-topic
  mqtthandle fun;      // callback function 
}
subscribe_type; 
subscribe_type mqtt_sub[MAX_MQTT_SUB];
int mqtt_sub_count=0;

String MQTT_Rx_Payload = "" ;
//--------- mqtt callback function 
void mqttcallback(char* to, byte* pay, unsigned int len) {
  String topic   = String(to);
  String payload = String((char*)pay);
  MQTT_Rx_Payload=payload.substring(0,len);
  Serial.println("\ncallback topic:" + topic + ", payload:" + MQTT_Rx_Payload);
  for (int i=0;i<mqtt_sub_count;i++) { // durchsuche alle subscriptions, bis topic passt 
    if (topic==mqtt_sub[i].topic) 
      mqtt_sub[i].fun(pay,len);         // Aufruf der richtigen callback-Funktion
  }
}

//------------ reconnect mqtt-client
void mqttreconnect() { // Loop until we're reconnected 
  if (!mqttclient.connected()) { 
    while (!mqttclient.connected()) { 
      Serial.print("Attempting MQTT connection...");
      if (mqttclient.connect("Oct1" , "user", "passwort" )) {
        Serial.println("connected");
        for (int i=0;i<mqtt_sub_count;i++) { // subscribe topic
          mqttclient.subscribe(mqtt_sub[i].topic.c_str());
          Serial.println("\nsubscribe");
          Serial.print(mqtt_sub[i].topic);
        }
      } 
      else { 
        Serial.print("failed, rc=");
        Serial.print(mqttclient.state());
        Serial.println(" try again in 5 seconds");
        delay(5000);
      }
    }
  } 
  else { 
    mqttclient.loop(); 
  }
}

void mqtt_callback_topic_start(byte* pay, unsigned int len){ // ---------- my callbackfunction mqtt
  String payload = String((char*)pay); // payload als String interpretieren
  MQTT_Rx_Payload=payload.substring(0,len);    // mit Länge von len Zeichen
  Serial.println("\n in callback payload:" + MQTT_Rx_Payload +"len:"+String(len));
  if (controlByte == 0){
    start();
  }
}

//IsrTimer für Taktung Aufnahme
void record()
{
  timer1_isr_init();
  timer1_attachInterrupt(myIsrTimer);
  timer1_enable(1,0,1); 
  timer1_write((clockCyclesPerMicrosecond() / 16) * MICRODELAY);
}

//Byteweise Erfassung Messwerte - 1 int16 in 2 byte
void myIsrTimer()
{
  wert = (int)analogRead(A0);
  dataCollection[currSample++] |= byte(wert >> 8);
  dataCollection[currSample++] |= byte(wert & 0x00FF);
  if (currSample == (SAMPLES*2)-2)
  {
    // only one more tick (0 at end is "one time" - timer mode)
    timer1_enable(1,0,0);
  }
}

//MQTT - Steuerung Audio Ausgabe RPi
void mqttSendSoundCommand() 
{
  String pay = "x";
  mqttreconnect();
  Serial.println("Start Ton abspielen");
  mqttclient.publish("PlaySound",pay.c_str());
  Serial.println("Sent");
}

//MQTT - Senden Messwerte, Aufnahme in MQTT-Pakete aufteilen (PubSubClient.h max. MQTT Größe ändern: #define MQTT_MAX_PACKET_SIZE 1024)
void mqttSendRecord() 
{
  mqttreconnect();
  for(int i = 0; i<SAMPLES*2/MQTTPACSIZE; i++){
    for(int j = 0; j < MQTTPACSIZE; j++){
      mqttpack[j] = dataCollection[(i*MQTTPACSIZE)+j];
    }
    mqttclient.publish("SensorData",mqttpack,MQTTPACSIZE);
  }
}

//MQTT - Steuerung Messdatenformatierung
void mqttSendFormatCommand() 
{
  String pay = "x";
  mqttreconnect();
  mqttclient.publish("FormatOutput",pay.c_str());
  Serial.println("Ende Daten formatieren");
}

//Setze Bytes zurück
void setZero(){
  for(int i=0;i<SAMPLES*2;i++){
    dataCollection[i] = 0;
  }
}

void start(){
  delayMicroseconds(100000);
  for (int i = 0; i < 1; i++){
    //RPi spiele Audio Datei ab (Node Red Kommando)
    controlByte = 1;
    mqttSendSoundCommand();
    delay(100);
    currSample = 0;
    //Starte Aufnahme
    Serial.println("Start record");
    record();
    delay(1200);
    //Sende Messdaten
    Serial.println("Ende record");
    mqttSendRecord();
    delay(50);
    //RPi formatiere Messdaten 
    mqttSendFormatCommand();
    delay(500);
    //Setze Bytes zurück
    setZero();
    controlByte = 0;
  }
}

void setup(){ // Einmalige Initialisierung
  Serial.begin(115200);
  //------------ WLAN initialisieren 
  WiFi.persistent(false);
  WiFi.mode(WIFI_STA);
  delay(100);
  Serial.print ("\nWLAN connect to:");
  Serial.print(ssid);
  WiFi.begin(ssid,password);
  while (WiFi.status() != WL_CONNECTED) { // Warte bis Verbindung steht 
    delay(500); 
    Serial.print(".");
  };
  Serial.println ("\nconnected, meine IP:"+ WiFi.localIP().toString());
  
  //----------------------------------MQTT-Client 
  mqttclient.setServer(mqtt_server, 1883);
  mqttclient.setCallback(mqttcallback);

  //--------- prepare mqtt subscription 
  mqtt_sub_count++; // add new element 
  if (mqtt_sub_count < MAX_MQTT_SUB) { 
    mqtt_sub[mqtt_sub_count-1].topic = "start";
    mqtt_sub[mqtt_sub_count-1].fun = mqtt_callback_topic_start; //callback function
  } 
  else Serial.println(" err max. mqtt subscription");

  mqttreconnect();
  controlByte = 0;
}

void loop(){
  mqttclient.loop();
  
  if (digitalRead(2)==LOW && controlByte == 0)
  {
   start();
  }
  else
  {
    controlByte = 0;
    delayMicroseconds(60);
  }
}
